package apiserver

import (
	"context"
	"fmt"
	"log"
	"net"

	"github.com/emicklei/go-restful"
	"github.com/google/k8s-stateless-subresource/pkg/internal/apiserver"
	"github.com/spf13/pflag"
	"go.uber.org/zap"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	"k8s.io/apimachinery/pkg/runtime"
	"k8s.io/apimachinery/pkg/runtime/serializer"
	genericapiserver "k8s.io/apiserver/pkg/server"
	"k8s.io/apiserver/pkg/server/options"
	"k8s.io/client-go/rest"
	"knative.dev/pkg/injection"
	"knative.dev/pkg/injection/sharedmain"
	"knative.dev/pkg/logging"
)

// Config has configuration for the API server.
type Config struct {
	// SecureServing has the configuration for securing the API server.
	SecureServing *options.SecureServingOptionsWithLoopback

	// Authentication has the configuration for authenticating the API server.
	Authentication *options.DelegatingAuthenticationOptions

	// Authorization has the configuration for authorization the API server.
	Authorization *options.DelegatingAuthorizationOptions

	// Features has the configuration for profiling the API server.
	Features *options.FeatureOptions

	// RESTConfig is the rest.Config used to communicate with the Kubernetes
	// API server. It can be nil.
	RESTConfig *rest.Config
}

// NewConfig returns a new config with the defaults set.
func NewConfig() Config {
	var cfg Config
	cfg.setDefaults()
	return cfg
}

// AddFlags adds flags for the various fields in the config. This is an
// optional (albeit convenient) way to populate the config.
func (c *Config) AddFlags(fs *pflag.FlagSet) {
	c.SecureServing.AddFlags(fs)
	c.Authentication.AddFlags(fs)
	c.Authorization.AddFlags(fs)
	c.Features.AddFlags(fs)
}

func (c *Config) setDefaults() {
	if c.RESTConfig == nil {
		env := injection.Flags()
		var err error
		if c.RESTConfig, err = injection.GetRESTConfig(
			env.ServerURL,
			env.Kubeconfig,
		); err != nil {
			log.Fatalf("failed to get rest config: %v", err)
		}
	}
	if c.SecureServing == nil {
		c.SecureServing = options.NewSecureServingOptions().WithLoopback()
	}
	if c.Authentication == nil {
		c.Authentication = options.NewDelegatingAuthenticationOptions()
	}
	if c.Authorization == nil {
		c.Authorization = options.NewDelegatingAuthorizationOptions()
	}
	if c.Features == nil {
		c.Features = options.NewFeatureOptions()
	}
}

func (c *Config) server(
	name string,
	codecs serializer.CodecFactory,
) (*apiserver.Server, error) {
	// Generate self-signed certs to use.
	if err := c.SecureServing.MaybeDefaultWithSelfSignedCerts(
		"localhost",
		nil,
		[]net.IP{net.ParseIP("127.0.0.1")},
	); err != nil {
		return nil, fmt.Errorf("failed to create self-signed certificates: %v", err)
	}

	serverConfig := genericapiserver.NewConfig(codecs)

	// Apply SecureServing config.
	if err := c.SecureServing.ApplyTo(
		&serverConfig.SecureServing,
		&serverConfig.LoopbackClientConfig,
	); err != nil {
		return nil, fmt.Errorf("failed to merge SecureServing config: %v", err)
	}

	// Apply Authentication config.
	if err := c.Authentication.ApplyTo(
		&serverConfig.Authentication,
		serverConfig.SecureServing,
		nil,
	); err != nil {
		return nil, fmt.Errorf("failed to merge Authentication config: %v", err)
	}

	// Apply Authorization config.
	if err := c.Authorization.ApplyTo(&serverConfig.Authorization); err != nil {
		return nil, fmt.Errorf("failed to merge Authorization config: %v", err)
	}

	// Don't bother with the informers.
	completedConfig := serverConfig.Complete(nil)

	genericServer, err := completedConfig.New(
		name,
		genericapiserver.NewEmptyDelegate(),
	)
	if err != nil {
		return nil, err
	}

	s := &apiserver.Server{
		GenericAPIServer: genericServer,
	}

	return s, nil
}

// EndpointConstructor is used to construct an endpoint.
type EndpointConstructor interface {
	// API returns the endpoint's API.
	API() (*runtime.Scheme, metav1.APIResource)

	// RegisterWebService is used to setup a WebService for the given
	// endpoint.
	RegisterWebService(context.Context, *restful.WebService) error
}

// EndpointConstructorStruct implements EndpointConstructor.
type EndpointConstructorStruct struct {
	// Scheme is the endpoint's Scheme.
	Scheme *runtime.Scheme

	// Resource is the endpoint's Resource.
	Resource metav1.APIResource

	// RegisterWebServiceF is used to setup a WebService for the given
	// endpoint.
	RegisterWebServiceF func(context.Context, *restful.WebService) error
}

// API implements EndpointConstructor.
func (f EndpointConstructorStruct) API() (*runtime.Scheme, metav1.APIResource) {
	return f.Scheme, f.Resource
}

// RegisterWebService implements EndpointConstructor.
func (f EndpointConstructorStruct) RegisterWebService(ctx context.Context, ws *restful.WebService) error {
	return f.RegisterWebServiceF(ctx, ws)
}

// Main invokes MainWithConfig with a nil config. It will block until the
// given context is done.
func Main(
	ctx context.Context,
	name string,
	ctor EndpointConstructor,
) {
	MainWithConfig(ctx, name, Config{}, ctor)
}

// MainWithConfig runs the generic main flow for API servers. It will block
// until the given context is done.
func MainWithConfig(
	ctx context.Context,
	name string,
	cfg Config,
	ctor EndpointConstructor,
) {
	// Set config defaults.
	cfg.setDefaults()

	// Setup injection.
	ctx = sharedmain.EnableInjectionOrDie(ctx, cfg.RESTConfig)

	// Setup logging.
	logger, atomicLevel := sharedmain.SetupLoggerOrDie(ctx, name)
	defer logger.Sync()
	ctx = logging.WithLogger(ctx, logger)

	sharedmain.CheckK8sClientMinimumVersionOrDie(ctx, logger)
	cmw := sharedmain.SetupConfigMapWatchOrDie(ctx, logger)
	sharedmain.WatchLoggingConfigOrDie(ctx, cmw, logger, atomicLevel, name)

	logger.Info("Starting configuration manager...")
	if err := cmw.Start(ctx.Done()); err != nil {
		logger.Fatalw("failed to start configuration manager", zap.Error(err))
	}

	scheme, resource := ctor.API()

	// Setup codecs.
	codecs := serializer.NewCodecFactory(scheme)

	s, err := cfg.server(name, codecs)
	if err != nil {
		logger.Fatalw("failed to setup API Server", zap.Error(err))
	}

	// Build a register function with the context as the first argument. The
	// underlying library doesn't really use/care about a context.
	register := func(ws *restful.WebService) error {
		return ctor.RegisterWebService(ctx, ws)
	}

	// Install the endpoint.
	if err := s.InstallAPI(
		scheme,
		codecs,
		resource,
		register,
	); err != nil {
		logger.Fatalw("failed to install API", zap.Error(err))
	}

	// Run the server.
	if err := s.GenericAPIServer.PrepareRun().Run(ctx.Done()); err != nil {
		logger.Fatalw("failed to run", zap.Error(err))
	}
}
